/************************************************************************/
/*  msextract -	Extract events from MiniSEED data files based on origin	*/
/*		and station location.					*/
/*		Extraction criteria based on phase arrival times and/or	*/
/*		origin times and duration.				*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Berkeley Seismological Laboratory				*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: msextract.c,v 1.9 2004/03/16 00:11:45 doug Exp $ ";
#endif

#include <stdlib.h>
#include <math.h>
#include <stdio.h>

#include "qlib2.h"
#include "version.h"
#include "externals.h"

#ifndef	DEFAULT_STATION_COORD_FILE
#define DEFAULT_STATION_COORD_FILE	"/usr/contrib/data/bdsn/stat.db.coord"
#endif

#ifndef	PRE_P_TIME
#define	PRE_P_TIME	30
#endif

#ifndef	POST_S_TIME
#define	POST_S_TIME	60
#endif

#ifndef	PRE_ORIGIN_TIME	
#define	PRE_ORIGIN_TIME	0
#endif

#ifndef	ORIGIN_DURATION
#define	ORIGIN_DURATION	300
#endif

#ifndef	WINDOW_SELECTION
#define WINDOW_SELECTION    "pf"
#endif

#define	PRINT_EVENT(info) { \
    if (win->event) \
        fprintf (info, "Event %s ", time_to_str(win->event->time,date_fmt)); \
    else \
        fprintf (info, "Box "); \
    fprintf (info, "(%s to ", time_to_str(win->start,date_fmt)); \
    fprintf (info, "%s) ", time_to_str(win->end,date_fmt)); \
}
char *syntax[] = {
"%s version " VERSION,
"%s  -e event_file [-b box_file] [-p time_span] [-s time_span]",
"   [-f time_span] [-d time_span] [-w window] [-o output_file]",
"   [-c coord_file] [-d n] [-h] [input_file_1]",
"    where:",
"	-e event_file",
"		    Required file containing list of events.",
"	-b box_file",
"		    Optional file with lat-lon box and start-end window.",
"		    If present, box window is added to window_selection.",
"	-p time_span",
"		    Time_span for pre_event data before P-wave.",
"		    Default is " STRING(PRE_EVENT_TIME) "S.",
"	-s time_span",
"		    Time_span for event duration after S-wave.",
"		    Default is " STRING(POST_EVENT_TIME) "S.",
"	-f time_span",
"		    Pre-origin start time.",
"		    Default is " STRING(PRE_ORIGIN_TIME) "S.",
"	-l time_span",
"		    Duration of time from pre-origin start time.",
"		    Default is " STRING(ORIGIN_DURATION) "S.",
"	-w window_selection",
"		    Specified window selection.",
"		    Can by any combination of the following letters:",
"		    p - phase (P and S) arrival critera.",
"		    f - fixed origin time criteria.",
"		    Default is " WINDOW_SELECTION ".",
"	-o outfile  Optional name of output file.",
"		    If not specified, write to stdout.", 
"	-c coord_file",
"		    Alternate station coordinate file.",
"		    Default file is " DEFAULT_STATION_COORD_FILE ,
"	-d n	    Debug option N.",
"		    1 = debug time window.",
"		    2 = debug save data records.",
"		    4 = debug discarded data records.",
"	-h	    Help - prints syntax message.",
"	input_file",
"		    Data file with original data.  If no file is specified,",
"		    read from stdin.",
NULL };


/************************************************************************/
/*  main:   main program						*/
/************************************************************************/
main (int argc, char **argv)
{
    EVENT *evts;
    FILE *fpin, *fpout, *fpev;
    DATA_HDR *hdr = NULL;
    INT_TIME e0, e1;
    WINDOW *box_window, *windows, *win;
    VERTEX *points;
    double w1,w2,tdif,dist_km;
    int iwin, nevts, n_box_points, nwindows, i, n, found;
    int left, num_data_bytes, blksize, status;
    int span_initialized = 0;
    char *mseed = NULL;
    char station[8];
    char *p, c;
    int data_fmt = MONTHS_FMT_1;

    /*	Initialization.							*/
    output_file= NULL;		/* empty initial output directory.	*/
    fpin = stdin;		/* assume input to come from stdin.	*/
    info = stdout;
    box_file = NULL;
    date_fmt = JULIAN_FMT;
    pre_p = PRE_P_TIME;
    post_s = POST_S_TIME;
    origin_duration = ORIGIN_DURATION;
    coord_file = DEFAULT_STATION_COORD_FILE;
    window_selection = WINDOW_SELECTION;

    parse_cmdline (&argc, &argv);

    if (event_file) {
	if ((fpev = fopen (event_file,"r")) == NULL) {
	    fprintf (info, "Unable to open event file %s\n", event_file);
	    exit(1);
	}
    }
    else {
	fprintf (info, "Missing event file name\n");
	exit(1);
    }

    if (input_file) {
	if ((fpin = fopen (input_file,"r")) == NULL) {
	    fprintf (info, "Unable to open input file %s\n", input_file);
	    exit(1);
	}
    }
    else {
	input_file = "<stdin>";
	fpin = stdin;
    }

    if (output_file) {
	info = stderr;
	if ((fpout = fopen (output_file,"w")) == NULL) {
	    fprintf (info, "Unable to open output file %s\n", output_file);
	    exit(1);
	}
    }
    else {
	output_file = "<stdout>";
	fpout = stdout;
    }

    nevts = get_events (fpev, &evts);
    if (nevts < 0) exit (1);
    
    blksize = EOF;
    if (box_file) n_box_points = get_box(&points, &box_window);
    if ((blksize = read_ms_record (&hdr, &mseed, fpin)) < 0 || hdr == NULL ) 
	exit(0);
    get_windows(evts, nevts, points, n_box_points, &windows, &nwindows, 
		box_window, hdr);

    /* read_ms_record will allocate buffer if mseed == NULL.		*/
    /* Re-use the buffer for subsequent records from the same file.	*/

    iwin = 0;
    found = 1;
    win = NULL;
    while (iwin < nwindows) {
	win = &windows[iwin];
	found = 0;
	
	/* We want continuous data from win start through win end.	*/
	while ((tdif = tdiff(win->start,hdr->endtime)) >= 0.) {
	    /* Skip this block -- too early for this event. */
	    if (debug(DEBUG_DATA_REJECT)) {
		fprintf (info, "Discard %s ", time_to_str(hdr->begtime,date_fmt));
		fprintf (info, "to %s\n", time_to_str(hdr->endtime,date_fmt));
		fflush (info);
	    }
	    free_data_hdr(hdr);
	    if ((blksize = read_ms_record (&hdr, &mseed, fpin)) < 0) break;
	}

	while (hdr && ((tdif = tdiff(win->end,hdr->begtime)) > 0.)) {
	    if (!found) {
		found = 1;
		PRINT_EVENT(info)
		fprintf (info, "- FOUND\n");
		fflush (info);
		e0 = hdr->begtime;
	    }
	    /* Save this block - it intersects the event.	*/
	    if (debug(DEBUG_DATA_SAVE)) {
		fprintf (info, "Saving  %s ", time_to_str(hdr->begtime,date_fmt));
		fprintf (info, "to %s\n", time_to_str(hdr->endtime,date_fmt));
		fflush (info);
	    }
	    if (fwrite (mseed, blksize, 1, fpout) != 1) {
		fprintf(info, "Error writing data block\n");
		exit(1);
	    }
	    e1 = hdr->endtime;
	    free_data_hdr(hdr);
	    if ((blksize = read_ms_record (&hdr, &mseed, fpin)) < 0) break;
	}
	if (! found) {
	    PRINT_EVENT(info)
	    fprintf (info, "- NOT FOUND\n");
	    fflush(info);
	    ++iwin;
	}
	else {
	    fprintf (info, "====>\tSave event %s ", time_to_str(e0, date_fmt));
	    fprintf (info, "to %s\n", time_to_str(e1, date_fmt));
	    fflush(info);
	    ++iwin;
	}
	if (! hdr) break;
    }

    if (blksize < EOF) {
	fprintf(info, "Error %d reading input file\n", blksize);
	exit(1);
    }

    /* Display any remaining events */
    while (iwin < nwindows) {
	win = &windows[iwin++];
	if (win->event)
	    fprintf (info, "Want event %s ", time_to_str(win->event->time,
							 date_fmt));
	else
	    fprintf(info, "Box window ");
	fprintf (info, "- NOT FOUND\n");
	fflush(info);
    }

    /* Cleanup before exit. */
    if (hdr) free_data_hdr;
    if (mseed) free (mseed);
    fclose (fpev);
    fclose (fpin);
    fclose (fpout);
    return(0);
}
